package rabinizer.exec;

import java.io.*;
import java.util.*;
import java.io.StringReader;

import rabinizer.formulas.*;
import rabinizer.parser.LTLParser;

/**
 * Entry point.
 * 
 * @author Jan Kretinsky & Zuzana Komarkova & Andreas Gaiser & Ruslan Ledesma-Garza
 *
 * @param <P>
 */
public class KRabinizer {


	public static void errorMessageAndExit(String text) {
		System.out.println("ERROR: " + text);
		System.exit(1);
	}

	public static void printUsage() {
		System.out.println(
				"\n"+
						"Usage: rabinizer [options] <ltlfile>\n"+
						"   Read LTL[F,G,X,U]-formula from file <ltlfile> and\n"+
						"   compute the corresponding Gen Rabin automaton\n" +
						"   and the corresponding Rabin automaton.\n"+
						"   By default, dump each constructed item to a corresponding\n"+
						"   file <ltlfile.*>.\n" +
						"Options:\n"+
						"   -h            : print this help and exit\n"+
						"   -v            : verbose terminal output\n"+
						"   -nodump       : only construct, do not dump results to files\n"+
						"   -nodot        : (deprecated) same as -nodump\n"+
						"   -ltl2dstar    : dump formula in reverse polish notation and exit\n\n"+
						"   -gen-edges    : creates generalized rabin automaton with condition on transitions\n"+
						"   -gen-states   : creates generalized rabin automaton with condition on states\n"+
						"   -rabin-edges  : creates rabin automaton with condition on transitions\n"+
						"   -rabin-states : creates rabin automaton with condition on states\n"+
						"   -all          : creates both generalized and simple rabin automaton with conditions both on edges and states\n\n"+
						"   -dotty-notBDD : full states, not BDD edges, not BDD acceptance condition\n"+
						"   -dotty-full   : full states, BDD edges, BDD acceptance condition\n"+
						"   -dotty        : shorten numbered states, BDD edges, BDD acceptance condition\n"+
						"   -numerical    : ltl2dstar format \n"+
                                                "   -hoa          : HOA (Hanoi omega-automata) format \n\n"+
						"   -formula      : reads the input formula as an argument\n"+
						"   -batch        : batch processing of more on formula per line in file\n");
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) throws IOException {

		System.out.println(
				        "\n******************************************************************************\n" +
						"* Rabinizer 3 by Jan Kretinsky and Zuzana Komarkova                          *\n" +
						/*"******************************************************************************\n" +
"* The constructor of deterministic Rabin automata for LTL-formulae           *\n" +
"* Release: Fri Feb 7 11:50:56 CEST 2013                                      *\n" +
"* Reused code of versions of Rabinizer                                       *\n" +
"* Version 1 by Andreas Gaiser                                                *\n" +
"* Version 2 by Ruslan Ledesma-Garza                                          *\n" +
"* Version 3 by Jan Kretinsky                                                 *\n" +
"* Version 4 by Zuzana Komarkova                                              *\n" +
						 */"******************************************************************************");


		// Phase 0: Parsing of arguments
		String ltlfile = null;
		String formula = null;
		boolean dump = true;
		boolean ltl2dstar = false;
		boolean genEdges = false;
		boolean genStates = false;
		boolean rabinEdges = false;
		boolean rabinStates = false;
		boolean form = false;
		Integer format = 0;
		System.out.print("Arguments:");
		for (String arg : args) {
			System.out.print(" "+arg);
			if (arg.equals("-h") ||
					arg.equals("--h") || 
					arg.equals("-help") ||
					arg.equals("--help") ) {
				printUsage();
				System.exit(0);
			} else if (arg.equals("-v") ||
					arg.equals("--v") || 
					arg.equals("-verbose") ||
					arg.equals("--verbose") )              Misc.verbose = true;
			else if (arg.equals("-nodump") ||
					arg.equals("--nodump") ||
					arg.equals("-nodot") ||
					arg.equals("--nodot"))         dump = false;
			else if (arg.equals("-batch") ||
					arg.equals("--batch"))          main2(args);
			else if (arg.equals("-ltl2dstar") ||
					arg.equals("--ltl2dstar"))     ltl2dstar = true;
			else if (arg.equals("-gen-edges") ||
					arg.equals("--gen-edges"))     genEdges = true;
			else if (arg.equals("-gen-states") ||
					arg.equals("--gen-states"))     genStates = true;
			else if (arg.equals("-rabin-edges") ||
					arg.equals("--rabin-edges"))     rabinEdges = true;
			else if (arg.equals("-rabin-states") ||
					arg.equals("--rabin-states"))     rabinStates = true;
			else if (arg.equals("-all") ||
					arg.equals("--all"))    { genEdges = true; genStates=true; rabinEdges=true; rabinStates=true;}
			else if (arg.equals("-numerical") ||
					arg.equals("--numerical"))     format = 2;
			else if (arg.equals("-dotty") ||
					arg.equals("--dotty"))      format = 1;
			else if (arg.equals("-dotty-full") ||
					arg.equals("--dotty-full"))     format = 0;
			else if (arg.equals("-dotty-notBDD") ||
					arg.equals("--dotty-notBDD"))     format = 3;
                        else if (arg.equals("-hoa") ||
					arg.equals("--hoa"))     format = 4;
			else if (arg.equals("-formula") ||
					arg.equals("--formula"))   form=true  ;
			else if (arg.substring(0, 1).equals("-")) {
				System.out.println("\n\nERROR: unknown option " + arg);
				printUsage();
				System.exit(1);
			}
				
			else   if(form) {
                            formula=arg;
                            ltlfile = "output";
                        }
			else ltlfile = arg;
		}
		if(!genEdges && !genStates && !rabinEdges && !rabinStates) genEdges=true;
		
		System.out.println("");
		if (ltlfile == null & !form) errorMessageAndExit("No input file was given.");

		// Phase 1: Parsing of the input formula
		Misc.verboseln("\nPhase 1: Parsing of the input formula\n");
		long start_time_parsing = System.currentTimeMillis();
		LTLParser parser = null;
		if(form){
			try { parser = new LTLParser(new StringReader(formula)); }
			catch (Exception e) {
				errorMessageAndExit(
						"Exception when opening the input file: "
								+ e.getLocalizedMessage()); }
			
		}
		else{
			try { parser = new LTLParser(new FileReader(new File(ltlfile))); }
			catch (Exception e) {
				errorMessageAndExit(
						"Exception when opening the input file: "
								+ e.getLocalizedMessage()); }
		}
		Formula phi = null;
		try { phi = parser.parse(); }
		catch (Exception e) {
			errorMessageAndExit("Exception when parsing: "
					+ e.getLocalizedMessage()); }
		System.out.println("Input formula: " + phi.toString());
		long interval_parsing = System.currentTimeMillis() - start_time_parsing;

		// Phase 1a: Dump ltl2dstar formula
		if (ltl2dstar) {
			System.out.println("ltl2dstar format: "+phi.toReversePolishString());
			System.exit(0);
		}

		// Phase 1b: Dump NNF formula
		phi = phi.toNNF();
		System.out.println("Input formula in NNF: "+phi);

		// Phase 1c: Check that the formula is in the fragment
		/*if (phi.untilInsideForall())
        	errorMessageAndExit("The formula is not in the fragment: there is an occurrence of until operator\n" +
        			"in a parameter of always operator.");
		 */

		// Phase 2: Enumeration of valuations
		Misc.verboseln("\nPhase 2: Enumeration of valuations");
		long start_time_valuations = System.currentTimeMillis();
		Misc.initializeValuations(Misc.bijectionIdAtom.size());
		BDDService.init(Misc.bijectionIdAtom.size());
		long interval_valuations = System.currentTimeMillis() - start_time_valuations;

		// Phases 3 & 4: segment automata & dfa
		long start_time_segment_automata, interval_segment_automata;
		long start_time_dfa, interval_dfa;

		/*
        // Go either for the old method 
        // - with one step history and
        // - with one state per valuation
        switch(method) {
       	case 0:
        	// Phase 3a: Construction of history automaton
       		start_time_segment_automata = System.currentTimeMillis();
	        List<SegmentAutomaton> historyAutomaton = new ArrayList<SegmentAutomaton>(1);
	        SegmentAutomaton hSA = new SegmentAutomaton(Misc.valuations);
	        historyAutomaton.add(hSA);
	        // Phase 3b: Construction of initial segment state
	        List<Set<Formula>> initialSegmentState = new ArrayList<Set<Formula>>(1);
	        Set<Formula> initSS = new HashSet<Formula>();
	        initSS.add(hSA.initialState);
	        initialSegmentState.add(initSS);
	        interval_segment_automata = System.currentTimeMillis() - start_time_segment_automata;	        

	        // Phase 4: Construction of DFA A
	        start_time_dfa = System.currentTimeMillis();
	        dfaA = new DFA(phi, historyAutomaton, initialSegmentState, Misc.valuations);
	        interval_dfa = System.currentTimeMillis() - start_time_dfa;

	    break;
        // ...or for the old method 
        // - with one step history and
        // - with collapsing of states...
       	case 1:
	        throw new Error("Rabinizer.main(): method 1 NYI");
        // break;

        // ...or the new method with segment automata
       	default:
		 */

		// Phase 3: Construction of segment automata
		Misc.verboseln("\n\nPhase 3: Construction of segment automata");

		// Phase 3a: Construction of always- and eventually- recurrent formulas aRF1, ..., aRFm, eRF1, ..., eRFn
		start_time_segment_automata = System.currentTimeMillis();
		//Tuple<Set<Formula>, Set<Formula>> recurrentFormulas = phi.recurrentFormulas();
		List<Formula> alwaysRecFs = new ArrayList<Formula>();
		alwaysRecFs.addAll(phi.argumentsAlways()); //(recurrentFormulas.left);
		//List<Formula> eventuallyRecFs = new ArrayList<Formula>();
		//eventuallyRecFs.addAll(recurrentFormulas.right);
		// Phase 3b: Construction of always segment automata aSA1, ..., aSAm and their corresponding initial segment states
		List<SegmentAutomaton> alwaysSAs = new ArrayList<SegmentAutomaton>(alwaysRecFs.size());
		List<Set<Formula>> alwaysISSs = new ArrayList<Set<Formula>>(alwaysRecFs.size());
		for (Formula aRF : alwaysRecFs) {
			// segment automaton
			//Misc.verboseln("\nAlways segment automaton");
			SegmentAutomaton aSA = new SegmentAutomaton(aRF, Misc.valuations);
			alwaysSAs.add(aSA);
			// corresponding initial segment state
			Set<Formula> iSS = new HashSet<Formula>();
			if (aSA.finalStates.contains(Misc.fTrue)) {
				iSS.add(Misc.fTrue);
			} else {
				boolean pickedFS = false;
				for(Formula f : aSA.finalStates) {
					if (!f.equals(Misc.fFalse)) {
						iSS.add(f);
						pickedFS = true;
						break;
					}
				}
				if (!pickedFS) iSS.add(Misc.fFalse);
			}
			Misc.verboseln("(always) Final state for iSS: "+iSS);
			iSS = aSA.backwardsReachability(iSS);
			Misc.verboseln("(always) Initial segment state: "+iSS);
			alwaysISSs.add(iSS);
		}
		// Phase 3c: Construction of eventually segment automata eSA1, ..., eSAm and their corresponding initial segment states
		/*
	        List<SegmentAutomaton> eventuallySAs = new ArrayList<SegmentAutomaton>(eventuallyRecFs.size());
	        List<Set<Formula>> eventuallyISSs = new ArrayList<Set<Formula>>(eventuallyRecFs.size());
	        for (Formula eRF : eventuallyRecFs) {
	        	// segment automaton
	        	Misc.verboseln("\nEventually segment automaton");
	        	SegmentAutomaton eSA = new SegmentAutomaton(eRF, Misc.valuations);
	        	eventuallySAs.add(eSA);
	        	/// corresponding initial segment state
	        	Set<Formula> iSS = new HashSet<Formula>();
	        	if (eSA.finalStates.contains(Misc.fFalse)) {
	        		iSS.add(Misc.fFalse);
	        	} else {
	        		boolean pickedFS = false;
	        		for(Formula f : eSA.finalStates) {
	        			if (!f.equals(Misc.fTrue)) {
	        				iSS.add(f);
	        				pickedFS = true;
	        				break;
	        			}
	        		}
	        		if (!pickedFS) iSS.add(Misc.fTrue);
	        	}
	        	Misc.verboseln("(always) Final state for iSS: "+iSS);
	        	iSS = eSA.backwardsReachability(iSS);
	        	eventuallyISSs.add(iSS);
	        	Misc.verboseln("(always) Initial segment state: "+iSS);
        	}
		 */     
		interval_segment_automata = System.currentTimeMillis() - start_time_segment_automata;

		// Phase 4: Construction of DFA A
		Misc.verboseln("\n\nPhase 4: Construction of Generalized Rabin Automaton");
		start_time_dfa = System.currentTimeMillis();
		//alwaysSAs.addAll(eventuallySAs);
		List<SegmentAutomaton> segmentAutomata = alwaysSAs;
		//alwaysISSs.addAll(eventuallyISSs);
		Misc.verboseln("");
		KDFA dfaA = new KDFA(phi, segmentAutomata, new KState(alwaysSAs).initialize(), Misc.valuations);
		KDFA dfaA2 = new KDFA(phi, segmentAutomata, new KState(alwaysSAs).initializeFilled(), Misc.valuations);
		if(dfaA2.states.size() < dfaA.states.size()){
			System.out.println("\n\nInitial states were optimized, saving " + (dfaA.states.size()-dfaA2.states.size()) + " states");
			dfaA = dfaA2;
		}else
			System.out.println("\n\nStandard initial states were used");

		interval_dfa = System.currentTimeMillis() - start_time_dfa;
		System.out.println("\nTransition-based Generalized Rabin Automaton: ");
		System.out.println("\tNumber of states    : " + dfaA.states.size());
		System.out.println("\tNumber of edges     : " + dfaA.edges.size());
		/*	        
        }
		 */

		// Phase 5: Construction of Generalized Rabin Acceptance Condition
		Misc.verboseln("\n\nPhase 5: Construction of transition-based Generalized Rabin Acceptance Condition");
		long start_time_grac = System.currentTimeMillis();
		Misc.verboseln("");
		List<Formula> subformulasAlways = new ArrayList<Formula>();
		subformulasAlways.addAll(phi.argumentsAlways());
		//List<Formula> subformulasExist = new ArrayList<Formula>();
		//subformulasExist.addAll(phi.argumentsFinsideG(false));
		Misc.verboseln("G-subformulas: "+subformulasAlways);

		// set numbered choice

		GeneralizedRabinAcceptanceCondition genRabinAC = new GeneralizedRabinAcceptanceCondition(subformulasAlways, dfaA, format);
		long interval_grac = System.currentTimeMillis() - start_time_grac;
		System.out.println("\tSize of Transition-based Generalized Rabin AC: " + genRabinAC.size());

		
    	// Phase 6: Construction of the Rabin Automaton
		KRabinAutomaton rabin=null;
		KRabinAcceptanceCondition rabinAC=null;
		KRabinAutomatonStates rabinS=null;
		KRabinStatesAcceptanceCondition rabinSAC=null;
		long interval_rabin=0;
		long interval_rabin_ac=0;
		long interval_rabinS=0;
		long interval_rabinS_ac=0;
		
		if(rabinEdges || rabinStates){
	    	Misc.verboseln("Phase 6: Rabin Automaton");
	    	long start_time_rabin = System.currentTimeMillis();
	    	rabin = new KRabinAutomaton(dfaA, genRabinAC);
	    	interval_rabin = System.currentTimeMillis() - start_time_rabin;
	        System.out.println("\n\nTransition-based Rabin Automaton: ");
	        System.out.println("\tNumber of states    : " + rabin.states.size());
	        System.out.println("\tNumber of edges     : " + rabin.edges.size());

    	// Phase 7: Construction of Rabin Acceptance Condition
	        
	    	Misc.verboseln("");
	    	Misc.verboseln("Phase 7: Construction of transition-based Rabin Acceptance Condition");
	    	long start_time_rabin_ac = System.currentTimeMillis();
	    	rabinAC = new KRabinAcceptanceCondition(rabin);
	    	interval_rabin_ac = System.currentTimeMillis() - start_time_rabin_ac;
	        System.out.println("\tSize of Transition-based Rabin AC    : " + rabinAC.size());
		
        
        // Phase 8: Construction of the state-based Rabin Automaton
	        if(rabinStates){
		    	Misc.verboseln("Phase 8: state-based Rabin Automaton");
		    	long start_time_rabinS = System.currentTimeMillis();
		    	rabinS = new KRabinAutomatonStates(rabin, rabinAC);
		    	interval_rabinS = System.currentTimeMillis() - start_time_rabinS;
		        System.out.println("\n\nState-based Rabin Automaton: ");
		        System.out.println("\tNumber of states    : " + rabinS.states.size());
		        System.out.println("\tNumber of edges     : " + rabinS.edges.size());

    	// Phase 9: Construction of the state-based Rabin Acceptance Condition
		    	Misc.verboseln("");
		    	Misc.verboseln("Phase 9: Construction of the state-based Rabin Acceptance Condition");
		    	long start_time_rabinS_ac = System.currentTimeMillis();
		    	rabinSAC = new KRabinStatesAcceptanceCondition(rabinS);
		    	interval_rabinS_ac = System.currentTimeMillis() - start_time_rabinS_ac;
		        System.out.println("\tSize of State-based Rabin AC    : " + rabinSAC.size());
	        }
		}
        
        // Phase 10: Construction of the state-based Generalized Rabin Automaton
		KDFAStates kdfaS=null;
		GeneralizedRabinStatesAcceptanceCondition grabinSAC=null;
		long interval_kdfaS=0;
		long interval_grabinS_ac=0;
		if(genStates){
	    	Misc.verboseln("\nPhase 10: state-based Generalized Rabin Automaton");
	    	long start_time_kdfaS = System.currentTimeMillis();
	    	kdfaS = new KDFAStates(dfaA, genRabinAC);
	    	interval_kdfaS = System.currentTimeMillis() - start_time_kdfaS;
	        System.out.println("\n\nState-based Generalized Rabin Automaton: ");
	        System.out.println("\tNumber of states    : " + kdfaS.states.size());
	        System.out.println("\tNumber of edges     : " + kdfaS.edges.size());

    	//Phase 11: Construction of the state-based Generalized Rabin Acceptance Condition
	    	Misc.verboseln("");
	    	Misc.verboseln("Phase 11: Construction of the state-based Generalized Rabin Acceptance Condition");
	    	long start_time_grabinS_ac = System.currentTimeMillis();
	    	grabinSAC = new GeneralizedRabinStatesAcceptanceCondition(kdfaS);
	    	interval_grabinS_ac = System.currentTimeMillis() - start_time_grabinS_ac;
	        System.out.println("\tSize of State-based Generalized Rabin AC    : " + grabinSAC.size());
		}
        

        // Phase 12: Dumping of files
                Misc.verboseln("\n\n Phase 12: Dumping files");
                long start_time_output = System.currentTimeMillis();
                if (dump && (format!=4)) {
                    if (genEdges) {
                        try {
                            FileWriter fw = new FileWriter(new File(ltlfile + ".grabin.dot"));
                            System.out.println("\n");
                            System.out.println("Dumping Transition-based Generalized Rabin Automaton");
                            if (format == 1) {
                                fw.write(dfaA.toDottyNumbered());
                            } else if (format == 0) {
                                fw.write(dfaA.toDotty());
                            } else if (format == 2) {
                                fw.write(dfaA.toDottyNumbered());
                                FileWriter fw2 = new FileWriter(new File(ltlfile + ".grabinNumerical.txt"));
                                fw2.write(dfaA.toNumerical(genRabinAC));
                                fw2.close();
                            } else if (format == 3) {
                                fw.write(dfaA.toDottyNotBDD());
                            }
                            fw.close();
                        } catch (IOException e) {
                            errorMessageAndExit("IO exception when writing file " + ltlfile + ".grabin.dot: " + e.getMessage());
                        }

                        try {
                            FileWriter fw = new FileWriter(new File(ltlfile + ".grac"));
                            //System.out.println("\n");
                            System.out.println("Dumping Transition-based Generalized Rabin Acceptance Condition");
                            fw.write(genRabinAC.toString());
                            fw.close();
                        } catch (IOException e) {
                            errorMessageAndExit("IO exception when writing file " + ltlfile + ".grac: " + e.getMessage());
                        }
                    }
                    if (rabinEdges) {
                        try {
                            FileWriter fw = new FileWriter(new File(ltlfile + ".rabin.dot"));
                            System.out.println("\n");
                            System.out.println("Dumping Transition-based Rabin Automaton");
                            fw.write(rabin.toDotty());
                            fw.close();
                        } catch (IOException e) {
                            errorMessageAndExit("IO exception when writing file " + ltlfile + ".rabin.dot: " + e.getMessage());
                        }
                        try {
                            FileWriter fw = new FileWriter(new File(ltlfile + ".rac"));
                            //System.out.println("\n");
                            System.out.println("Dumping Transition-based Rabin Acceptance Condition");
                            fw.write(rabinAC.toString());
                            fw.close();
                        } catch (IOException e) {
                            errorMessageAndExit("IO exception when writing file " + ltlfile + ".rac: " + e.getMessage());
                        }
                    }
                    if (rabinStates) {
                        try {
                            FileWriter fw = new FileWriter(new File(ltlfile + ".rabinStates.dot"));
                            System.out.println("\n");
                            System.out.println("Dumping State-based Rabin Automaton");
                            fw.write(rabinS.toDotty());
                            fw.close();
                        } catch (IOException e) {
                            errorMessageAndExit("IO exception when writing file " + ltlfile + ".rabin.dot: " + e.getMessage());
                        }
                        try {
                            FileWriter fw = new FileWriter(new File(ltlfile + ".rsac"));
                            //System.out.println("\n");
                            System.out.println("Dumping State-based Rabin Acceptance Condition");
                            fw.write(rabinSAC.toString());
                            fw.close();
                        } catch (IOException e) {
                            errorMessageAndExit("IO exception when writing file " + ltlfile + ".rac: " + e.getMessage());
                        }
                    }
                    if (genStates) {
                        try {
                            FileWriter fw = new FileWriter(new File(ltlfile + ".gRabinStates.dot"));
                            System.out.println("\n");
                            System.out.println("Dumping State-based Generalized Rabin Automaton");
                            fw.write(kdfaS.toDotty());
                            fw.close();
                        } catch (IOException e) {
                            errorMessageAndExit("IO exception when writing file " + ltlfile + ".rabin.dot: " + e.getMessage());
                        }
                        try {
                            FileWriter fw = new FileWriter(new File(ltlfile + ".grsac"));
                            //System.out.println("\n");
                            System.out.println("Dumping State-based Generalized Rabin Acceptance Condition");
                            System.out.println("\n");
                            fw.write(grabinSAC.toString());
                            fw.close();
                        } catch (IOException e) {
                            errorMessageAndExit("IO exception when writing file " + ltlfile + ".rac: " + e.getMessage());
                        }
                    }

                    int i = 0;
                    Misc.verboseln("\n");
                    if (genEdges || genStates) {
                        for (SegmentAutomaton sA : segmentAutomata) {
                            try {
                                FileWriter fw = new FileWriter(new File(ltlfile + ".segment" + (++i) + ".dot"));
                                //System.out.println("");
                                System.out.println("Dumping Segment Automaton #" + i);
                                fw.write(sA.toDotty());
                                fw.close();
                            } catch (IOException e) {
                                errorMessageAndExit("IO exception when writing file " + ltlfile + ".segment" + i + ".dot:" + e.getMessage());
                            }
                        }
                    }
                } else if (dump && (format == 4)) {
                    try {
                        FileWriter fw = new FileWriter(new File(ltlfile + ".hoa"));
                        System.out.println("\n");

                        if (genEdges) {
                            System.out.println("Dumping Transition-based Generalized Rabin Automaton");
                            fw.write(dfaA.toHOAF(genRabinAC));
                        }
                        if (rabinEdges) {
                            System.out.println("Dumping Transition-based Rabin Automaton");
                            fw.write(rabin.toHOAF(rabinAC));
                        }
                        if (rabinStates) {
                            System.out.println("Dumping State-based Rabin Automaton");
                            fw.write(rabinS.toHOAF(rabinSAC));
                        }
                        if (genStates) {
                            System.out.println("Dumping State-based Generalized Rabin Automaton");
                            fw.write(kdfaS.toHOAF(grabinSAC));
                            System.out.println("\n");
                        }

                        fw.close();
                    } catch (IOException e) {
                        errorMessageAndExit("IO exception when writing file " + ltlfile + ".hoa: " + e.getMessage());
                    }
                }
                long interval_output = System.currentTimeMillis() - start_time_output;


		// End of the program
		System.out.println("\n\nSUMMARY\n******************************************************************************");
		if(genEdges && !(genStates && rabinEdges && rabinStates)){
			System.out.println("\nTransition-based Generalized Rabin Automaton: ");
			System.out.println("\tNumber of states    : " + dfaA.states.size());
			System.out.println("\tNumber of edges     : " + dfaA.edges.size());
			System.out.println("\tSize of AC          : " + genRabinAC.size());
			System.out.println("\n");
		}
		
		if(!(genStates && genEdges && rabinStates) && rabinEdges){
			System.out.println("\nTransition-based Rabin Automaton: ");
			System.out.println("\tNumber of states    : " + rabin.states.size());
			System.out.println("\tNumber of edges     : " + rabin.edges.size());
			System.out.println("\tSize of AC          : " + rabinAC.size());
			System.out.println("\n");
		}
		
		if(!(genStates && genEdges && rabinEdges) && rabinStates){
			System.out.println("\nState-based Rabin Automaton: ");
			System.out.println("\tNumber of states    : " + rabinS.states.size());
			System.out.println("\tNumber of edges     : " + rabinS.edges.size());
			System.out.println("\tSize of AC          : " + rabinAC.size());
			System.out.println("\n");
		}
		
		if(!(rabinEdges && genEdges && rabinStates) && genStates){
			System.out.println("\nState-based Generalized Rabin Automaton: ");
			System.out.println("\tNumber of states    : " + kdfaS.states.size());
			System.out.println("\tNumber of edges     : " + kdfaS.edges.size());
			System.out.println("\tSize of AC          : " + grabinSAC.size());
			System.out.println("\n");
		}
		
		
		System.out.println(
				"Construction time for Transition-based GRA : "
						+ (interval_parsing
								+ interval_valuations 
								+ interval_segment_automata
								+ interval_dfa
								+ interval_grac
								//     		       + interval_rabin
								//        		   + interval_rabin_ac
								)
								+ "msec");
		if (dump) 
			System.out.println(
				"Total time (incl. output to files)         : "
						+ (interval_parsing
								+ interval_valuations 
								+ interval_segment_automata
								+ interval_dfa
								+ interval_grac
								//              		   + interval_rabin
								//              		   + interval_rabin_ac
								+ interval_output)
								+ "msec");
		
			
		System.out.println(
				"   Parsing of input formula                : "
						+ interval_parsing + "msec");
		System.out.println(
				"   Enumeration of valuations               : "
						+ interval_valuations + "msec");
		System.out.println(
				"   Construction of segment automata        : "
						+ interval_segment_automata + "msec");
		System.out.println(
				"   Construction of DFA                     : "
						+ interval_dfa + "msec");
		System.out.println(
				"   Construction of GRA condition           : "
						+ interval_grac + "msec");
		
		
		//System.out.println("\n");
		if(rabinEdges || rabinStates){
			System.out.println(
					"Total time for Transition-based RA         : "
							+ (interval_parsing
									+ interval_valuations 
									+ interval_segment_automata
									+ interval_dfa
									+ interval_grac
									+ interval_rabin
								    + interval_rabin_ac
									) + "msec");
			
		}
		if(rabinStates){
			System.out.println(
					"Total time for State-based RA              : "
							+ (interval_parsing
									+ interval_valuations 
									+ interval_segment_automata
									+ interval_dfa
									+ interval_grac
									+ interval_rabin
								    + interval_rabin_ac
								    + interval_rabinS
								    + interval_rabinS_ac
									) + "msec");
			
		}
		if(genStates){
			System.out.println(
					"Total time for State-based GRA             : "
							+ (interval_parsing
									+ interval_valuations 
									+ interval_segment_automata
									+ interval_dfa
									+ interval_grac
									+ interval_kdfaS
								    + interval_grabinS_ac
									) + "msec");
		}
		
		/*
        System.out.println(
                           "Construction of Rabin Automaton            : "
                           + interval_rabin + "msec");
        System.out.println(
                           "Construction of Rabin Acceptance Condition : "
                           + interval_rabin_ac + "msec");
		 */
		//System.out.println("\n");
		if (dump) System.out.println(
				"Output to files                            : "
						+ interval_output + "msec");
                System.out.println("******************************************************************************\n");
		//Prism.tryout();

	}
	
	//for more formulas in one file
	public static void main2(String[] args) throws IOException {

		
		// Phase 0: Parsing of arguments
		String ltlfile = null;
		boolean dump = true;
		boolean ltl2dstar = false;
		boolean genEdges = false;
		boolean genStates = false;
		boolean rabinEdges = false;
		boolean rabinStates = false;
		Integer format = 0;
		//System.out.print("Arguments:");
		for (String arg : args) {
			System.out.print(" "+arg);
			if (arg.equals("-h") ||
					arg.equals("--h") || 
					arg.equals("-help") ||
					arg.equals("--help") ) {
				printUsage();
				System.exit(0);
			} else if (arg.equals("-v") ||
					arg.equals("--v") || 
					arg.equals("-verbose") ||
					arg.equals("--verbose") )              Misc.verbose = true;
			else if (arg.equals("-nodump") ||
					arg.equals("--nodump") ||
					arg.equals("-nodot") ||
					arg.equals("--nodot"))         dump = false;
			else if (arg.equals("-ltl2dstar") ||
					arg.equals("--ltl2dstar"))     ltl2dstar = true;
			else if (arg.equals("-batch") ||
					arg.equals("--batch"))          ;
			else if (arg.equals("-gen-edges") ||
					arg.equals("--gen-edges"))     genEdges = true;
			else if (arg.equals("-gen-states") ||
					arg.equals("--gen-states"))     genStates = true;
			else if (arg.equals("-rabin-edges") ||
					arg.equals("--rabin-edges"))     rabinEdges = true;
			else if (arg.equals("-rabin-states") ||
					arg.equals("--rabin-states"))     rabinStates = true;
			else if (arg.equals("-all") ||
					arg.equals("--all"))    { genEdges = true; genStates=true; rabinEdges=true; rabinStates=true;}
			else if (arg.equals("-numerical") ||
					arg.equals("--numerical"))     format = 2;
			else if (arg.equals("-dotty") ||
					arg.equals("--dotty"))      format = 1;
			else if (arg.equals("-dotty-full") ||
					arg.equals("--dotty-full"))     format = 0;
			else if (arg.equals("-dotty-notBDD") ||
					arg.equals("--dotty-notBDD"))     format = 3;
			else if (arg.substring(0, 1).equals("-")) {
				System.out.println("\n\nERROR: unknown option " + arg);
				printUsage();
				System.exit(1);
			} else                                     ltlfile = arg;
		}
		if(!genEdges && !genStates && !rabinEdges && !rabinStates) genEdges=true;
		
		System.out.println("");
		if (ltlfile == null) errorMessageAndExit("No input file was given.");
		
		if(genEdges && !(genStates && rabinEdges && rabinStates)){
			System.out.println("\nTransition-based Generalized Rabin Automaton, output in form: \n");
			System.out.println("Input formula");
			System.out.println("Number of states");
			System.out.println("Size of AC");
			
		}
		
		if(!(genStates && genEdges && rabinStates) && rabinEdges){
			System.out.println("\nTransition-based Rabin Automaton, output in form: \n");
			System.out.println("Input formula");
			System.out.println("Number of states");
			System.out.println("Size of AC");
			
		}
		
		if(!(genStates && genEdges && rabinEdges) && rabinStates){
			System.out.println("\nState-based Rabin Automaton, output in form: \n");
			System.out.println("Input formula");
			System.out.println("Number of states");
			System.out.println("Size of AC");
			
		}
		
		if(!(rabinEdges && genEdges && rabinStates) && genStates){
			System.out.println("\nState-based Generalized Rabin Automaton, output in form: \n");
			System.out.println("Input formula");
			System.out.println("Number of states");
			System.out.println("Size of AC");
			
			
		}
		System.out.println("__________________________________________________________________________");
		
		BufferedReader bReader=new BufferedReader(new FileReader(new File(ltlfile)));
		String s;
	 	while((s=bReader.readLine())!=null){
	 		
			// Phase 1: Parsing of the input formula
			Misc.verboseln("\nPhase 1: Parsing of the input formula\n");
			long start_time_parsing = System.currentTimeMillis();
			LTLParser parser = null;
			try { parser = new LTLParser(new StringReader(s)); }
			catch (Exception e) {
				errorMessageAndExit(
						"Exception when opening the input file: "
								+ e.getLocalizedMessage()); }
			Formula phi = null;
			try { phi = parser.parse(); }
			catch (Exception e) {
				errorMessageAndExit("Exception when parsing: "
						+ e.getLocalizedMessage()); }
			System.out.println(phi.toString());
			long interval_parsing = System.currentTimeMillis() - start_time_parsing;
	
			// Phase 1a: Dump ltl2dstar formula
			if (ltl2dstar) {
				//System.out.println("ltl2dstar format: "+phi.toReversePolishString());
				System.exit(0);
			}
	
			// Phase 1b: Dump NNF formula
			phi = phi.toNNF();
			//System.out.println("Input formula in NNF: "+phi);
	
			// Phase 1c: Check that the formula is in the fragment
			/*if (phi.untilInsideForall())
	        	errorMessageAndExit("The formula is not in the fragment: there is an occurrence of until operator\n" +
	        			"in a parameter of always operator.");
			 */
	
			// Phase 2: Enumeration of valuations
			Misc.verboseln("\nPhase 2: Enumeration of valuations");
			long start_time_valuations = System.currentTimeMillis();
			Misc.initializeValuations(Misc.bijectionIdAtom.size());
			BDDService.init(Misc.bijectionIdAtom.size());
			long interval_valuations = System.currentTimeMillis() - start_time_valuations;
	
			// Phases 3 & 4: segment automata & dfa
			long start_time_segment_automata, interval_segment_automata;
			long start_time_dfa, interval_dfa;
	
	
			// Phase 3: Construction of segment automata
			Misc.verboseln("\n\nPhase 3: Construction of segment automata");
	
			// Phase 3a: Construction of always- and eventually- recurrent formulas aRF1, ..., aRFm, eRF1, ..., eRFn
			start_time_segment_automata = System.currentTimeMillis();
			//Tuple<Set<Formula>, Set<Formula>> recurrentFormulas = phi.recurrentFormulas();
			List<Formula> alwaysRecFs = new ArrayList<Formula>();
			alwaysRecFs.addAll(phi.argumentsAlways()); //(recurrentFormulas.left);
			//List<Formula> eventuallyRecFs = new ArrayList<Formula>();
			//eventuallyRecFs.addAll(recurrentFormulas.right);
			// Phase 3b: Construction of always segment automata aSA1, ..., aSAm and their corresponding initial segment states
			List<SegmentAutomaton> alwaysSAs = new ArrayList<SegmentAutomaton>(alwaysRecFs.size());
			List<Set<Formula>> alwaysISSs = new ArrayList<Set<Formula>>(alwaysRecFs.size());
			for (Formula aRF : alwaysRecFs) {
				// segment automaton
				//Misc.verboseln("\nAlways segment automaton");
				SegmentAutomaton aSA = new SegmentAutomaton(aRF, Misc.valuations);
				alwaysSAs.add(aSA);
				// corresponding initial segment state
				Set<Formula> iSS = new HashSet<Formula>();
				if (aSA.finalStates.contains(Misc.fTrue)) {
					iSS.add(Misc.fTrue);
				} else {
					boolean pickedFS = false;
					for(Formula f : aSA.finalStates) {
						if (!f.equals(Misc.fFalse)) {
							iSS.add(f);
							pickedFS = true;
							break;
						}
					}
					if (!pickedFS) iSS.add(Misc.fFalse);
				}
				Misc.verboseln("(always) Final state for iSS: "+iSS);
				iSS = aSA.backwardsReachability(iSS);
				Misc.verboseln("(always) Initial segment state: "+iSS);
				alwaysISSs.add(iSS);
			}
			    
			interval_segment_automata = System.currentTimeMillis() - start_time_segment_automata;
	
			// Phase 4: Construction of DFA A
			Misc.verboseln("\n\nPhase 4: Construction of Generalized Rabin Automaton");
			start_time_dfa = System.currentTimeMillis();
			//alwaysSAs.addAll(eventuallySAs);
			List<SegmentAutomaton> segmentAutomata = alwaysSAs;
			//alwaysISSs.addAll(eventuallyISSs);
			Misc.verboseln("");
			KDFA dfaA = new KDFA(phi, segmentAutomata, new KState(alwaysSAs).initialize(), Misc.valuations);
			KDFA dfaA2 = new KDFA(phi, segmentAutomata, new KState(alwaysSAs).initializeFilled(), Misc.valuations);
			if(dfaA2.states.size() < dfaA.states.size()){
				//System.out.println("\n\nInitial states were optimized, saving " + (dfaA.states.size()-dfaA2.states.size()) + " states");
				dfaA = dfaA2;
			}else
				//System.out.println("\n\nStandard initial states were used");
	
			interval_dfa = System.currentTimeMillis() - start_time_dfa;
			//System.out.println("\nTransition-based Generalized Rabin Automaton: ");
			//System.out.println("\tNumber of states    : " + dfaA.states.size());
			//System.out.println("\tNumber of edges     : " + dfaA.edges.size());
			/*	        
	        }
			 */
	
			// Phase 5: Construction of Generalized Rabin Acceptance Condition
			Misc.verboseln("\n\nPhase 5: Construction of transition-based Generalized Rabin Acceptance Condition");
			long start_time_grac = System.currentTimeMillis();
			Misc.verboseln("");
			List<Formula> subformulasAlways = new ArrayList<Formula>();
			subformulasAlways.addAll(phi.argumentsAlways());
			//List<Formula> subformulasExist = new ArrayList<Formula>();
			//subformulasExist.addAll(phi.argumentsFinsideG(false));
			Misc.verboseln("G-subformulas: "+subformulasAlways);
	
			// set numbered choice
	
			GeneralizedRabinAcceptanceCondition genRabinAC = new GeneralizedRabinAcceptanceCondition(subformulasAlways, dfaA, format);
			long interval_grac = System.currentTimeMillis() - start_time_grac;
			//System.out.println("\tSize of Transition-based Generalized Rabin AC: " + genRabinAC.size());
	
			
	    	// Phase 6: Construction of the Rabin Automaton
			KRabinAutomaton rabin=null;
			KRabinAcceptanceCondition rabinAC=null;
			KRabinAutomatonStates rabinS=null;
			KRabinStatesAcceptanceCondition rabinSAC=null;
			long interval_rabin=0;
			long interval_rabin_ac=0;
			long interval_rabinS=0;
			long interval_rabinS_ac=0;
			
			if(rabinEdges || rabinStates){
		    	Misc.verboseln("Phase 6: Rabin Automaton");
		    	long start_time_rabin = System.currentTimeMillis();
		    	rabin = new KRabinAutomaton(dfaA, genRabinAC);
		    	interval_rabin = System.currentTimeMillis() - start_time_rabin;
		        //System.out.println("\n\nTransition-based Rabin Automaton: ");
		        //System.out.println("\tNumber of states    : " + rabin.states.size());
		        //System.out.println("\tNumber of edges     : " + rabin.edges.size());
	
	    	// Phase 7: Construction of Rabin Acceptance Condition
		        
		    	Misc.verboseln("");
		    	Misc.verboseln("Phase 7: Construction of transition-based Rabin Acceptance Condition");
		    	long start_time_rabin_ac = System.currentTimeMillis();
		    	rabinAC = new KRabinAcceptanceCondition(rabin);
		    	interval_rabin_ac = System.currentTimeMillis() - start_time_rabin_ac;
		        //System.out.println("\tSize of Transition-based Rabin AC    : " + rabinAC.size());
			
	        
	        // Phase 8: Construction of the state-based Rabin Automaton
		        if(rabinStates){
			    	Misc.verboseln("Phase 8: state-based Rabin Automaton");
			    	long start_time_rabinS = System.currentTimeMillis();
			    	rabinS = new KRabinAutomatonStates(rabin, rabinAC);
			    	interval_rabinS = System.currentTimeMillis() - start_time_rabinS;
			        //System.out.println("\n\nState-based Rabin Automaton: ");
			        //System.out.println("\tNumber of states    : " + rabinS.states.size());
			        //System.out.println("\tNumber of edges     : " + rabinS.edges.size());
	
	    	// Phase 9: Construction of the state-based Rabin Acceptance Condition
			    	Misc.verboseln("");
			    	Misc.verboseln("Phase 9: Construction of the state-based Rabin Acceptance Condition");
			    	long start_time_rabinS_ac = System.currentTimeMillis();
			    	rabinSAC = new KRabinStatesAcceptanceCondition(rabinS);
			    	interval_rabinS_ac = System.currentTimeMillis() - start_time_rabinS_ac;
			       // System.out.println("\tSize of State-based Rabin AC    : " + rabinSAC.size());
		        }
			}
	        
	        // Phase 10: Construction of the state-based Generalized Rabin Automaton
			KDFAStates kdfaS=null;
			GeneralizedRabinStatesAcceptanceCondition grabinSAC=null;
			long interval_kdfaS=0;
			long interval_grabinS_ac=0;
			if(genStates){
		    	Misc.verboseln("\nPhase 10: state-based Generalized Rabin Automaton");
		    	long start_time_kdfaS = System.currentTimeMillis();
		    	kdfaS = new KDFAStates(dfaA, genRabinAC);
		    	interval_kdfaS = System.currentTimeMillis() - start_time_kdfaS;
		        //System.out.println("\n\nState-based Generalized Rabin Automaton: ");
		        //System.out.println("\tNumber of states    : " + kdfaS.states.size());
		       // System.out.println("\tNumber of edges     : " + kdfaS.edges.size());
	
	    	//Phase 11: Construction of the state-based Generalized Rabin Acceptance Condition
		    	Misc.verboseln("");
		    	Misc.verboseln("Phase 11: Construction of the state-based Generalized Rabin Acceptance Condition");
		    	long start_time_grabinS_ac = System.currentTimeMillis();
		    	grabinSAC = new GeneralizedRabinStatesAcceptanceCondition(kdfaS);
		    	interval_grabinS_ac = System.currentTimeMillis() - start_time_grabinS_ac;
		        //System.out.println("\tSize of State-based Generalized Rabin AC    : " + grabinSAC.size());
			}
	        
	
			// Phase 12: Dumping of files
			Misc.verboseln("\n\n Phase 12: Dumping files");
			long start_time_output = System.currentTimeMillis();
			if (dump) {
				if(genEdges){
					try {
						FileWriter fw = new FileWriter(new File(ltlfile+".grabin.dot"));
						//System.out.println("\n");
						//System.out.println("Dumping Transition-based Generalized Rabin Automaton");
						if(format==1)fw.write(dfaA.toDottyNumbered());
						else if(format==0) fw.write(dfaA.toDotty());
						else if(format==2) {
							fw.write(dfaA.toDottyNumbered());
							FileWriter fw2 = new FileWriter(new File(ltlfile+".grabinNumerical.txt"));
							fw2.write(dfaA.toNumerical(genRabinAC));
							fw2.close();
						}
						else if(format==3){
							fw.write(dfaA.toDottyNotBDD());
						}
						fw.close();
					} catch (IOException e) {
						errorMessageAndExit("IO exception when writing file "+ltlfile+".grabin.dot: " + e.getMessage()); }
		
					try {
						FileWriter fw = new FileWriter(new File(ltlfile+".grac"));
						//System.out.println("\n");
						//System.out.println("Dumping Transition-based Generalized Rabin Acceptance Condition");
						fw.write(genRabinAC.toString());
						fw.close();
					} catch (IOException e) {
						errorMessageAndExit("IO exception when writing file "+ltlfile+".grac: " + e.getMessage()); }
				}
				if(rabinEdges){
			        try {
			            FileWriter fw = new FileWriter(new File(ltlfile+".rabin.dot"));
			            //System.out.println("\n");
			            //System.out.println("Dumping Transition-based Rabin Automaton");
			            fw.write(rabin.toDotty());
			            fw.close();
			        } catch (IOException e) {
			            errorMessageAndExit("IO exception when writing file "+ltlfile+".rabin.dot: " + e.getMessage()); }
		            try {
			            FileWriter fw = new FileWriter(new File(ltlfile+".rac"));
			            //System.out.println("\n");
			            //System.out.println("Dumping Transition-based Rabin Acceptance Condition");
			            fw.write(rabinAC.toString());
			            fw.close();
			        } catch (IOException e) {
			            errorMessageAndExit("IO exception when writing file "+ltlfile+".rac: " + e.getMessage()); }
				}
				if(rabinStates){
		            try {
			            FileWriter fw = new FileWriter(new File(ltlfile+".rabinStates.dot"));
			            //System.out.println("\n");
			            //System.out.println("Dumping State-based Rabin Automaton");
			            fw.write(rabinS.toDotty());
			            fw.close();
			        } catch (IOException e) {
			            errorMessageAndExit("IO exception when writing file "+ltlfile+".rabin.dot: " + e.getMessage()); }
		            try {
			            FileWriter fw = new FileWriter(new File(ltlfile+".rsac"));
			            //System.out.println("\n");
			            //System.out.println("Dumping State-based Rabin Acceptance Condition");
			            fw.write(rabinSAC.toString());
			            fw.close();
			        } catch (IOException e) {
			            errorMessageAndExit("IO exception when writing file "+ltlfile+".rac: " + e.getMessage()); }
				}
				if(genStates){
		            try {
			            FileWriter fw = new FileWriter(new File(ltlfile+".gRabinStates.dot"));
			            //System.out.println("\n");
			            //System.out.println("Dumping State-based Generalized Rabin Automaton");
			            fw.write(kdfaS.toDotty());
			            fw.close();
			        } catch (IOException e) {
			            errorMessageAndExit("IO exception when writing file "+ltlfile+".rabin.dot: " + e.getMessage()); }
		            try {
			            FileWriter fw = new FileWriter(new File(ltlfile+".grsac"));
			            //System.out.println("\n");
			            //System.out.println("Dumping State-based Generalized Rabin Acceptance Condition");
			            //System.out.println("\n");
			            fw.write(grabinSAC.toString());
			            fw.close();
			        } catch (IOException e) {
			            errorMessageAndExit("IO exception when writing file "+ltlfile+".rac: " + e.getMessage()); }
				}
				
				int i = 0;
				Misc.verboseln("\n");
				if(genEdges || genStates){
					for (SegmentAutomaton sA : segmentAutomata) {
						try {
							FileWriter fw = new FileWriter(new File(ltlfile+".segment"+(++i)+".dot"));
							//System.out.println("");
							//System.out.println("Dumping Segment Automaton #"+i);
							fw.write(sA.toDotty());
							fw.close();
						} catch (IOException e) {
							errorMessageAndExit("IO exception when writing file "+ltlfile+".segment"+i+".dot:" + e.getMessage()); }
					}
				}
			}
			long interval_output = System.currentTimeMillis() - start_time_output;
	
	
			// End of the program
			
			if(genEdges && !(genStates && rabinEdges && rabinStates)){
				//System.out.println("\nTransition-based Generalized Rabin Automaton: ");
				System.out.println(dfaA.states.size());
				//System.out.println("\tNumber of edges     : " + dfaA.edges.size());
				System.out.println(genRabinAC.size());
				
			}
			
			if(!(genStates && genEdges && rabinStates) && rabinEdges){
				//System.out.println("\nTransition-based Rabin Automaton: ");
				System.out.println(rabin.states.size());
				//System.out.println("\tNumber of edges     : " + rabin.edges.size());
				System.out.println(rabinAC.size());
				
			}
			
			if(!(genStates && genEdges && rabinEdges) && rabinStates){
				//System.out.println("\nState-based Rabin Automaton: ");
				System.out.println(rabinS.states.size());
				//System.out.println("\tNumber of edges     : " + rabinS.edges.size());
				System.out.println(rabinAC.size());
				
			}
			
			if(!(rabinEdges && genEdges && rabinStates) && genStates){
				//System.out.println("\nState-based Generalized Rabin Automaton: ");
				System.out.println(kdfaS.states.size());
				//System.out.println("\tNumber of edges     : " + kdfaS.edges.size());
				System.out.println(grabinSAC.size());
				
			}
			
	        
			//Prism.tryout();
		 }
	 	bReader.close();
		System.exit(0);

	}
}
